# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2019-05-30 07:10
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Webhook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True, db_index=True)),
                ('target_url', models.CharField(max_length=255)),
                ('is_active', models.BooleanField(default=True)),
                ('secret_key', models.CharField(blank=True, max_length=124, null=True)),
                ('event_type', models.CharField(max_length=124)),
                ('callback', models.CharField(blank=True, max_length=64, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='WebhookEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True, db_index=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, unique=True)),
                ('request_payload', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('response_content', models.TextField()),
                ('response_http_status', models.IntegerField()),
                ('delivered', models.BooleanField(default=False)),
                ('request_datetimes', django.contrib.postgres.fields.ArrayField(base_field=models.DateTimeField(blank=True, null=True), default=list, size=None)),
                ('webhook', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='whisperer.Webhook')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterUniqueTogether(
            name='webhook',
            unique_together=set([('user', 'target_url', 'event_type')]),
        ),
    ]



from rest_framework import serializers

from whisperer.tests.models import Order


class OrderSerializer(serializers.ModelSerializer):
    class Meta:
        model = Order
        fields = '__all__'



from whisperer.exceptions import WebhookAlreadyRegistered
from whisperer.models import Webhook


class WebhookService(object):
    def register_webhook(self, user, *args, **kwargs):
        event_type = kwargs.get('event_type')
        target_url = kwargs.get('target_url')
        try:
            Webhook.objects.get(user=user, target_url=target_url, event_type=event_type)
            raise WebhookAlreadyRegistered()
        except Webhook.DoesNotExist:
            pass
        webhook = Webhook(user=user)
        for attr, value in kwargs.items():
            setattr(webhook, attr, value)
        webhook.save()
        return webhook

    def update_webhook(self, webhook, user, *args, **kwargs):
        webhook.user = user
        target_url = kwargs.get('target_url')
        event_type = kwargs.get('event_type')
        try:
            Webhook.objects.get(user=user, target_url=target_url, event_type=event_type)
            raise WebhookAlreadyRegistered()
        except Webhook.DoesNotExist:
            pass
        for attr, value in kwargs.items():
            setattr(webhook, attr, value)
        webhook.save()
        return webhook

    def delete_webhook(self, webhook):
        webhook.is_active = False
        webhook.save(update_fields=['is_active'])



from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from whisperer.utils import Registry

countdown_kwargs_serializers = Registry()


@countdown_kwargs_serializers.register(key='fixed')
class FixedRetryCountdownKwargsSerializer(serializers.Serializer):
    seconds = serializers.IntegerField(min_value=0, write_only=True)


@countdown_kwargs_serializers.register(key='linear')
class LinearRetryCountdownKwargsSerializer(serializers.Serializer):
    base = serializers.IntegerField(min_value=0, write_only=True)
    limit = serializers.IntegerField(
        min_value=0, required=False, allow_null=True, write_only=True
    )


@countdown_kwargs_serializers.register(key='exponential')
class ExponentialRetryCountdownKwargsSerializer(serializers.Serializer):
    base = serializers.IntegerField(min_value=0, write_only=True)
    limit = serializers.IntegerField(
        min_value=0, required=False, allow_null=True, write_only=True
    )


@countdown_kwargs_serializers.register(key='random')
class RandomRetryCountdownKwargsSerializer(serializers.Serializer):
    min_value = serializers.IntegerField(min_value=0, write_only=True)
    max_value = serializers.IntegerField(min_value=0, write_only=True)

    def validate(self, attrs):
        attrs = super(RandomRetryCountdownKwargsSerializer, self).validate(attrs)
        if attrs['min_value'] >= attrs['max_value']:
            raise ValidationError("'max_value' must be greater than 'min_value'.")
        return attrs



from django.db.models.signals import post_save

from whisperer.events import WhispererEvent, registry
from whisperer.tasks import deliver_event
from whisperer.tests.models import Order
from whisperer.tests.serializers import OrderSerializer


class OrderCreateEvent(WhispererEvent):
    serializer_class = OrderSerializer
    event_type = 'order-created'


class OrderUpdateEvent(WhispererEvent):
    serializer_class = OrderSerializer
    event_type = 'order-updated'


registry.register(OrderCreateEvent)
registry.register(OrderUpdateEvent)


def signal_receiver(instance, created=False, **kwargs):
    if created:
        deliver_event(instance, 'order-created')
    else:
        deliver_event(instance, 'order-updated')


post_save.connect(signal_receiver, Order)



# Generated by Django 3.0.4 on 2020-04-01 08:47

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('whisperer', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='webhook',
            name='additional_headers',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



from __future__ import unicode_literals

from django.utils.module_loading import autodiscover_modules


def autodiscover():
    from whisperer.events import registry
    from whisperer.tasks import deliver_event

    autodiscover_modules('webhooks', register_to=registry)


default_app_config = 'whisperer.apps.WhispererConfig'



import uuid
from datetime import timedelta

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models
from django.utils import timezone

from whisperer.countdown import countdown_classes


class StarterModel(models.Model):
    created_date = models.DateTimeField(auto_now_add=True)
    modified_date = models.DateTimeField(auto_now=True, db_index=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        update_fields = kwargs.pop('update_fields', None)
        if update_fields is not None:
            if not isinstance(update_fields, list):
                update_fields = list(update_fields)
            update_fields.append('modified_date')
            kwargs['update_fields'] = update_fields
        super(StarterModel, self).save(*args, **kwargs)


class Webhook(StarterModel):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    target_url = models.CharField(max_length=255)
    is_active = models.BooleanField(default=True)
    secret_key = models.CharField(max_length=124, null=True, blank=True)
    event_type = models.CharField(max_length=124)
    callback = models.CharField(max_length=64, null=True, blank=True)
    retry_countdown_config = JSONField()
    additional_headers = JSONField(default=dict)
    config = JSONField(default=dict)

    def __str__(self):
        return '{}:{}'.format(self.event_type, self.target_url[:25])

    @property
    def countdown(self):
        config = self.retry_countdown_config
        countdown_class = countdown_classes[config['choice']]
        countdown_ = countdown_class(**config['kwargs'])
        return countdown_

    class Meta:
        unique_together = [('user', 'target_url', 'event_type')]


class WebhookEvent(StarterModel):
    uuid = models.UUIDField(default=uuid.uuid4, unique=True)
    webhook = models.ForeignKey(Webhook, on_delete=models.PROTECT)
    request_payload = JSONField(default=dict)
    response_content = models.TextField()
    response_http_status = models.IntegerField()
    delivered = models.BooleanField(default=False)
    request_datetimes = ArrayField(
        models.DateTimeField(null=True, blank=True), default=list
    )
    retry_count = models.PositiveIntegerField(null=True)

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey()

    def __str__(self):
        return '{}:{}'.format(self.webhook, self.delivered)

    @property
    def is_retry_allowed(self):
        countdown = self.webhook.countdown
        return (
            self.created_date + timedelta(seconds=countdown.get_value(self.retry_count))
        ) <= timezone.now()



import random

from whisperer.utils import Registry

countdown_classes = Registry()


class BaseRetryCountdown(object):
    def get_value(self, retry_count):
        raise NotImplementedError

    def __repr__(self):
        attrs = ['%s: %s' % (k, v) for k, v in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attrs))


@countdown_classes.register(key='fixed')
class FixedRetryCountdown(BaseRetryCountdown):
    def __init__(self, seconds):
        super(FixedRetryCountdown, self).__init__()
        self.seconds = seconds

    def get_value(self, retry_count):
        return self.seconds


@countdown_classes.register(key='linear')
class LinearRetryCountdown(BaseRetryCountdown):
    def __init__(self, base, limit=None):
        super(LinearRetryCountdown, self).__init__()
        self.base = base
        self.limit = limit

    def get_value(self, retry_count):
        count = self.base * retry_count
        if self.limit is None:
            return count
        return min(count, self.limit)


@countdown_classes.register(key='exponential')
class ExponentialRetryCountdown(BaseRetryCountdown):
    def __init__(self, base, limit=None):
        super(ExponentialRetryCountdown, self).__init__()
        self.base = base
        self.limit = limit

    def get_value(self, retry_count):
        count = self.base ** retry_count
        if self.limit is None:
            return count
        return min(count, self.limit)


@countdown_classes.register(key='random')
class RandomRetryCountdown(BaseRetryCountdown):
    def __init__(self, min_value, max_value):
        super(RandomRetryCountdown, self).__init__()
        self.min_value = min_value
        self.max_value = max_value

    def get_value(self, retry_count):
        return random.randint(self.min_value, self.max_value)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2019-05-30 07:10
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('whisperer', '0003_auto_20200505_0853'),
    ]

    operations = [
        migrations.AddField(
            model_name='WebhookEvent',
            name='retry_count',
            field=models.PositiveIntegerField(null=True),
        )
    ]



from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from whisperer.countdown import countdown_classes
from whisperer.models import Webhook, WebhookEvent
from whisperer.validators import countdown_kwargs_serializers


class RetryCountdownConfigSerializer(serializers.Serializer):
    choice = serializers.CharField()
    kwargs = serializers.JSONField()

    def validate_choice(self, choice):
        if not (choice in countdown_kwargs_serializers and choice in countdown_classes):
            raise ValidationError('"%s" is not a valid choice.' % choice)
        return choice

    def validate(self, attrs):
        serializer_class = countdown_kwargs_serializers[attrs['choice']]
        serializer = serializer_class(data=attrs['kwargs'])
        if not serializer.is_valid():
            raise ValidationError({'kwargs': serializer.errors})
        attrs['kwargs'] = serializer.validated_data
        return attrs


class WebhookSerializer(serializers.ModelSerializer):
    retry_countdown_config = RetryCountdownConfigSerializer()

    class Meta:
        model = Webhook
        exclude = ('user',)


class WebhookEventSerializer(serializers.ModelSerializer):
    webhook = WebhookSerializer()

    class Meta:
        model = WebhookEvent
        fields = '__all__'



from rest_framework import viewsets

from whisperer.models import Webhook, WebhookEvent
from whisperer.resources.filters import WebhookEventFilter, WebhookFilter
from whisperer.resources.serializers import WebhookEventSerializer, WebhookSerializer
from whisperer.services import WebhookService


class WebhookViewSet(viewsets.ModelViewSet):
    queryset = Webhook.objects.all()
    serializer_class = WebhookSerializer
    service = WebhookService()
    filter_class = WebhookFilter

    def get_queryset(self):
        queryset = super(WebhookViewSet, self).get_queryset()
        return queryset.filter(user=self.request.user)

    def perform_create(self, serializer):
        user = self.request.user
        serializer.instance = self.service.register_webhook(
            user, **serializer.validated_data
        )

    def perform_update(self, serializer):
        user = self.request.user
        self.service.update_webhook(
            serializer.instance, user=user, **serializer.validated_data
        )

    def perform_destroy(self, instance):
        self.service.delete_webhook(instance)


class WebhookEventViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = WebhookEvent.objects.all()
    filter_class = WebhookEventFilter
    serializer_class = WebhookEventSerializer

    def get_queryset(self):
        queryset = super(WebhookEventViewSet, self).get_queryset()
        return queryset.filter(webhook__user=self.request.user)
